
/*====================================

   Module: MOVIE
   Header-file (.HPP)
   (C) Copyright Greg Arakelian 1994,1995.
   All Rights Reserved.

  ===================================*/

#ifndef __STANDARD_MOVIE_HPP
#define __STANDARD_MOVIE_HPP


// ====================================
//  Include Files
// ====================================

//
// Standard
//

#ifndef __WING_H
#define __WING_H
#include "wing.h"
#endif

// ====================================
//  Miscellaneous
// ====================================

void ClearSystemPalette();
void CreateIdentityPalette (PALETTEENTRY *pe);


// ====================================
//  Autodesk Flic (FLC/FLI)
// ====================================

#define FLIHUGE far
#define FLISIZE WORD

typedef char FLIHUGE *LPPEL;

//
// Frame chunks contain the pixel and color data for the animation.
// A frame chunk may contain multiple subordinate chunks, each
// containing a different type of data for the current frame.  Each
// frame chunk starts with a 16-byte header that describes the
// contents of the frame:
//

typedef struct {
  int32 size;		// Size of flic including this header.
  uint16 type;		// Either FLI_TYPE or FLC_TYPE below.
  uint16 frames;	// Number of frames in flic, not including ring frame (Max 4000).
  uint16 width;		// Flic width in pixels.
  uint16 height;	// Flic height in pixels.
  uint16 depth;		// Bits per pixel.  (Always 8 now.)
  uint16 flags;		// FLI_FINISHED | FLI_LOOPED ideally.
  int32 speed;		// Delay between frames.
  short reserved1;	// Set to zero.
  uint32 created;	// Date of flic creation. (FLC only.)
  uint32 creator;	// Serial # of flic creator. (FLC only.) - 464C4942 ("FLIB")
  uint32 updated;	// Date of flic update. (FLC only.)
  uint32 updater;	// Serial # of flic updater. (FLC only.)
  uint16 aspect_dx;	// Width of square rectangle. (FLC only.)
  uint16 aspect_dy;	// Height of square rectangle. (FLC only.)
  char reserved2[38];	// Set to zero.
  int32 oframe1;	// Offset to frame 1. (FLC only.)
  int32 oframe2;	// Offset to frame 2. (FLC only.)
  char reserved3[40];	// Set to zero.
} FLICHEAD;


//
// Immediately following the frame header are the frame's subordinate
// data chunks.  When the chunks count in the frame header is zero, it
// indicates that this frame is identical to the previous frame.  This
// implies that no change is made to the screen or color palette, but
// the appropriate delay is still inserted during playback.
//

typedef struct {
  int32 size;		/* Size of frame including header. */
  uint16 type;		/* Always FRAME_TYPE */
  int16 chunks;		/* Number of subordinate chunks in frame. */
  char reserved[8];	/* Always 0. */
} FrameChunk;


// ====================================
//  TFlic CLASS DEFINITION
// ====================================

class TFlic
{
protected:
  enum {
    Success = 0,			// success
    ErrBadFlic = -1,			// bad flic file
    ErrOpen = -2,			// could not open flic
    ErrNoMemory = -3,			// not enough memory
    ErrBadFrame = -4,			// invalid frame data
    ErrRead = -5 };			// could not read flic
  TDib *FrameDib;			// TDib used for decompression
  bool IsResource;			// true if flic file is resource
  FLICHEAD Header;			// flic file header
  int CurrentFrame;			// number of current frame
  bool PaletteChanged;			// true if palette changed in decode
  int FlicFile;				// handle to flic file (if opened)
  HGLOBAL FlicImage;			// memory block holding flic file
  char HUGE *ImageBuf;			// pointer to flic file image (if any)
  char HUGE *ImageBufCurPos;		// pointer to location in flic file
  RGBQUAD FAR *FrameDibColors;		// pointer to TDib's color table
  LPPEL FrameDibBits;			// pointer to TDib's bitmap bits
  int WidthBytes;			// bytes per scan line in TDib
  int ErrorCode;			// error code from initialization
  int Initialize();
  int CalculateFrame2();
  void DecodeBlack (uint8 FLIHUGE *data);
  void DecodeBrun (uint8 FLIHUGE *data);
  void DecodeColor (uint8 FLIHUGE *data);
  void DecodeCopy (uint8 FLIHUGE *data);
  void DecodeLC (uint8 FLIHUGE *data);
  void DecodeSS2 (uint8 FLIHUGE *data);
  int DecodeFrame (FrameChunk *frame, uint8 FLIHUGE *data);
  LPPEL GetScanLine (int y) { return FrameDibBits + (FLISIZE)((long)(Header.height-y-1) * (long)WidthBytes); }
  int ReadBlock (void far *block, unsigned size);
  int ReadHugeBlock (char FLIHUGE *block, uint32 size);
  void Seek (long Position);
public:
  PALETTEENTRY PaletteEntry[256];
  TFlic (const char far *flicFilename, bool memoryLoad=false);
  TFlic (HINSTANCE hInstance, LPCSTR lpszName);
  virtual ~TFlic();
  int DecodeFrame (bool *lastFrame=NULL);
  inline bool GetCurrentFrame() { return CurrentFrame; }
  inline TDib *GetFrameDib() { return FrameDib; }
  inline TSize GetFrameSize() { return TSize(Header.width,Header.height); }
  inline bool GetPaletteChanged() { bool paletteChanged=PaletteChanged; PaletteChanged=false; return paletteChanged; }
  inline long GetSpeedMsec() { return Header.speed; }
  inline int GetErrorCode() { return ErrorCode; }
  inline bool IsOK() { return ErrorCode==0; }
  operator TDib &() { return *FrameDib; }
};


// ====================================
//  GetScanBytes (given width)
// ====================================

inline long GetScanBytes (int width)
{

  //
  // Assuming 8bpp, calculate size of scan line in bytes.
  // This is equal to (Pixel width * bits per pixel)
  // rounded up to a DWORD (32-bit) boundary.
  //
  return (((long)width*8+31) / 32) * 4;

}


// ====================================
//  GetScanBytes (given TDib)
// ====================================

inline long GetScanBytes (TDib &dib)
{

  //
  // Calculate and return size of scan line in bytes.
  // This is equal to (Pixel width * bits per pixel)
  // rounded up to a DWORD (32-bit) boundary.
  //
  int pixelWidth = dib.Width();
  int bitsPerPixel = dib.GetInfoHeader()->biBitCount;
  return (((long)pixelWidth*bitsPerPixel+31) / 32) * 4;

}


// ====================================
//  CopyDIBBits; TransCopyDIBBits
// ====================================

#ifndef WIN32
typedef uint8 huge *UHCHAR;
#else
typedef uint8 far *UHCHAR;
#endif


void PASCAL far CopyDIBBits (
  UHCHAR destBits,		// pointer to destination
  UHCHAR sourceBits,		// pointer to source
  short destScanBytes,		// size of destination scan line
  short sourceScanBytes,	// size of source scan line
  short width, short height);	// width and height of source

void PASCAL far TransCopyDIBBits (
  UHCHAR destBits,		// pointer to destination
  UHCHAR sourceBits,		// pointer to source
  short destScanBytes,		// size of destination scan line
  short sourceScanBytes,	// size of source scan line
  short width, short height,	// width and height of source
  uint8 transparentColor);	// transparent color


// ====================================
//  TWinGModule CLASS DEFINITION
// ====================================

#define WING_ERROR                              20000
#define WING_ERROR_WING_LOAD_FAILED		WING_ERROR+0
#define WING_ERROR_INCORRECT_WING		WING_ERROR+1

#ifdef WIN32
#define WINGModuleName "WING32.DLL"
#else
#define WINGModuleName "WING.DLL"
#endif

class TWinGModule
{
public:
  struct WINGPROCS {
    HDC (WINGAPI *pfnWinGCreateDC)(void);
    BOOL (WINGAPI *pfnWinGRecommendDIBFormat)(BITMAPINFO FAR *);
    HBITMAP (WINGAPI *pfnWinGCreateBitmap)(HDC, BITMAPINFO const FAR *, void FAR *FAR *);
    void FAR *(WINGAPI *pfnWinGGetDIBPointer)(HBITMAP, BITMAPINFO FAR *);
    UINT (WINGAPI *pfnWinGGetDIBColorTable)(HDC, UINT, UINT, RGBQUAD FAR *);
    UINT (WINGAPI *pfnWinGSetDIBColorTable)(HDC, UINT, UINT, RGBQUAD const FAR *);
    HPALETTE (WINGAPI *pfnWinGCreateHalftonePalette)(void);
    HBRUSH (WINGAPI *pfnWinGCreateHalftoneBrush)(HDC, COLORREF, WING_DITHER_TYPE);
    BOOL (WINGAPI *pfnWinGBitBlt)(HDC, int, int, int, int, HDC, int, int);
    BOOL (WINGAPI *pfnWinGStretchBlt)(HDC, int, int, int, int, HDC, int, int, int, int);
  } WinGProcs;
  TModule *Module;
  TWinGModule() { Module=0; }
  ~TWinGModule() { delete Module; Module=0; }
  void BindModule();
};

extern TWinGModule WinGModule;

#define WinGCreateDC			WinGModule.WinGProcs.pfnWinGCreateDC
#define WinGRecommendDIBFormat		WinGModule.WinGProcs.pfnWinGRecommendDIBFormat
#define WinGCreateBitmap		WinGModule.WinGProcs.pfnWinGCreateBitmap
#define WinGGetDIBPointer		WinGModule.WinGProcs.pfnWinGGetDIBPointer
#define WinGGetDIBColorTable		WinGModule.WinGProcs.pfnWinGGetDIBColorTable
#define WinGSetDIBColorTable		WinGModule.WinGProcs.pfnWinGSetDIBColorTable
#define WinGCreateHalftonePalette	WinGModule.WinGProcs.pfnWinGCreateHalftonePalette
#define WinGCreateHalftoneBrush		WinGModule.WinGProcs.pfnWinGCreateHalftoneBrush
#define WinGBitBlt			WinGModule.WinGProcs.pfnWinGBitBlt
#define WinGStretchBlt			WinGModule.WinGProcs.pfnWinGStretchBlt


// ====================================
//  TWinGDC CLASS DEFINITION
// ====================================

typedef struct {
  BITMAPINFOHEADER bmiHeader;
  RGBQUAD Colors[256];
} WINGBITMAPINFO;

class TWinGDC : public TCreatedDC
{
protected:
  TSize Size;
  int Orientation;
  void HUGE *FrameBits;
  WINGBITMAPINFO Info;
  HBITMAP WinGBitmap;
  HBITMAP OldMonoBitmap;
public:
  TWinGDC (TSize size, TPalette *palette=0);
  virtual ~TWinGDC();
  inline void HUGE *GetBits() { return FrameBits; }
  inline BITMAPINFO far *GetInfo() { return (BITMAPINFO far *)&Info; }
  inline int GetOrientation() { return Orientation; }
  inline TSize GetSize() { return Size; }
  void SetDIBColorTable (TPalette *palette=0);
  bool WinGDCBitBlt (const TDC &dstDC, int dstX, int dstY, int w, int h, int srcX, int srcY);
  bool WinGDCBitBlt (const TDC &dstDC, const TRect &dst, const TPoint &src);
  bool WinGDCStretchBlt (const TDC &dstDC, int dstX, int dstY, int dstW, int dstH, int srcX, int srcY, int srcW, int srcH);
  bool WinGDCStretchBlt (const TDC &dstDC, const TRect &dst, const TRect &src);
};


inline bool TWinGDC::WinGDCBitBlt (const TDC &dstDC, int dstX, int dstY, int w, int h, int srcX, int srcY)
{
  return WinGBitBlt(dstDC, dstX, dstY, w, h, GetHDC(), srcX, srcY);
}


inline bool TWinGDC::WinGDCBitBlt (const TDC &dstDC, const TRect &dst, const TPoint &src)
{
  return WinGBitBlt(dstDC, dst.left, dst.top, dst.Width(), dst.Height(), GetHDC(), src.x, src.y);
}


inline bool TWinGDC::WinGDCStretchBlt (const TDC &dstDC, int dstX, int dstY, int dstW, int dstH, int srcX, int srcY, int srcW, int srcH)
{
  return WinGStretchBlt(dstDC, dstX, dstY, dstW, dstH, GetHDC(), srcX, srcY, srcW, srcH);
}


inline bool TWinGDC::WinGDCStretchBlt(const TDC &dstDC, const TRect &dst, const TRect &src)
{
  return WinGStretchBlt(dstDC, dst.left, dst.top, dst.Width(), dst.Height(), GetHDC(), src.left, src.top, src.Width(), src.Height());
}


// ====================================
//  TRenderBuffer CLASS DEFINITION
// ====================================

class TSprite;  // forward declaration

class TRenderBuffer
{
protected:
  TWinGDC *WinGDC;		// WinGDC (when WinG bound dynamically)
  TDib *FrameDib;		// only used without WinG
  TSize FrameSize;		// size of animation frame
  void HUGE *FrameBits;		// pointer to frame dib's bits
  int Orientation;		// top-down or bottom-down orientation?
  int ScanBytes;		// scan bytes per line
  TPalette *Palette;		// palette to use for rendering
  TSprite *PaletteOwner;	// sprite which owns palette
  bool PaletteChanged;		// true if palette has changed
public:
  enum TOrientation { BottomUp=1, TopDown=-1 };
  TRenderBuffer (TSize frameSize);
  virtual ~TRenderBuffer();
  inline bool CanSetPalette (TSprite *sprite)
    { return !PaletteOwner || (PaletteOwner==sprite); }
  void ClearBits();
  void ClearPaletteChanged() { PaletteChanged=false; }
  void HUGE *GetBits() { return FrameBits; }
  TDib *GetDib() { return FrameDib; }
  int GetOrientation() { return Orientation; }
  TPalette *GetPalette() { return Palette; }
  TSprite *GetPaletteOwner() { return PaletteOwner; }
  bool GetPaletteChanged() { return PaletteChanged; }
  TRect GetRect() { return TRect(TPoint(0,0),FrameSize); }
  int GetScanBytes() { return ScanBytes; }
  TSize GetSize() { return FrameSize; }
  TWinGDC *GetWinGDC() { return WinGDC; }
  inline void ReleasePalette (TSprite *sprite)
    { if (sprite==PaletteOwner) SetPalette(0,0); }
  void SetPalette (TPalette *palette, TSprite *paletteOwner);
};


// ====================================
//  TSprite CLASS DEFINITION
// ====================================

class TSpriteGroup;

class TSprite
{
public:
  enum TRenderMode {
    Opaque = 0,			// all bits copied to frame
    Transparent = 1 };		// bits copied with transparency
  enum TRenderState {
    Stopped = 0,		// sprite is stopped (not advanced or rendered)
    Playing = 1,		// sprite is playing (both advanced and rendered)
    Paused = 2 };		// sprite is paused (not advanced but rendered)
protected:
  friend TSpriteGroup;
  TSpriteGroup *Group;		// group to which sprite belongs
  TSprite *NextSprite;		// link to next sprite
  TSprite *PrevSprite;		// link to previous sprite
  TPoint Position;		// current sprite position
  TPoint StartPosition;		// original sprite position
  TSize Size;			// sprite size
  TSize Motion;			// delta movement per frame
  uint8 TransparentColor;	// transparent color index
  TRenderMode RenderMode;	// rendering mode
  TRenderState RenderState;	// rendering state
  bool Ending;			// causing rendering to stop with sprite
  bool Condemned;		// should be deleted?
  bool AutoCondemn;		// condemn when stopped?
public:
  TSprite();
  ~TSprite() { Discard(); }
  virtual void Advance (DWORD deltaTime=0);
  void Center (TSize frameSize);
  inline void Condemn() { Condemned=true; }
  virtual void Discard() {}
  inline TSprite *GetNextSprite() { return NextSprite; }
  inline TPoint GetPosition() { return Position; }
  inline TSize GetSize() { return Size; }
  inline TRenderState GetRenderState() { return RenderState; }
  inline bool IsCondemned() { return Condemned; }
  inline bool IsEnding() { return Ending; }
  inline bool IsPaused() { return RenderState==Paused; }
  inline bool IsPlaying() { return RenderState==Playing; }
  inline bool IsStopped() { return RenderState==Stopped; }
  virtual TDib *LockDib (TRenderBuffer &renderBuffer) { return 0; }
  inline void Pause() { SetRenderState(Paused); }
  inline void Play() { SetRenderState(Playing); }
  virtual void Render (TRenderBuffer &renderBuffer);
  virtual void Reset() { Position=StartPosition; SetRenderState(Playing); }
  inline void SetAutoCondemn (bool autoCondemn) { AutoCondemn=autoCondemn; }
  inline void SetEnding (bool ending) { Ending=ending; }
  inline void SetMotion (TSize motion) { Motion=motion; }
  inline void SetPosition (TPoint position) { Position=StartPosition=position; }
  inline void SetRenderMode (TRenderMode renderMode) { RenderMode=renderMode; }
  virtual void SetRenderState (TRenderState renderState);
  inline void SetTransparentColor (uint8 transparentColor) { TransparentColor=transparentColor; }
  inline void Stop() { SetRenderState(Stopped); }
};


// ====================================
//  TSpriteGroup CLASS DEFINITION
// ====================================

class TSpriteGroup : public TSprite
{
protected:
  TSprite *FirstSprite;		// first sprite in group
  bool Continuous;		// restart group if stopped
  bool RenderedAll;		// rendered last item in group?
  TSprite *LastRendered;	// last sprite rendered in group
  void DeleteAll();
public:
  TSpriteGroup() : TSprite() { FirstSprite=0; Continuous=0; }
  ~TSpriteGroup() { DeleteAll(); }
  void Advance (DWORD deltaTime=0);
  TSprite *Append (TSprite *sprite, bool ending=false);
  inline bool GetContinuous() { return Continuous; }
  TSprite *Prepend (TSprite *sprite, bool ending=false);
  void Render (TRenderBuffer &renderBuffer);
  void Remove (TSprite *Sprite);
  virtual void Reset();
  void SetRenderState (TRenderState renderState);
  inline void SetContinuous (bool continuous) { Continuous=continuous; }
};


// ====================================
//  TDibSprite CLASS DEFINITION
// ====================================

class TDibSprite : public TSprite
{
protected:
  TDib *Dib;			// Sprite dib
public:
  TDibSprite (TDib *dib=0) : TSprite() { Dib=0; SetDib(dib); }
  ~TDibSprite() { delete Dib; Dib=0; }
  virtual TDib *LockDib (TRenderBuffer &renderBuffer) { return Dib; }
  virtual void Render (TRenderBuffer &renderBuffer);
  virtual void SetDib (TDib *dib) { delete Dib; if ((Dib=dib)!=0) Size=Dib->Size(); }
};


// ====================================
//  TFlicSprite CLASS DEFINITION
// ====================================

class TFlicSprite : public TSprite
{
protected:
  TFlic *Flic;			// pointer to flic
  char far *FlicFilename;	// flic filename (for reloads)
  bool MemoryLoad;		// load flic into memory for speed
  int NumLoops;			// number of times to loop
  int LoopsLeft;		// number of loops remaining
  TFlic *LoadFlic();
public:
  TFlicSprite (const char far *flicFilename, bool memoryLoad=false, int numLoops=0);
  TFlicSprite (TModule *module, LPCSTR flicResId, int numLoops=0);
  ~TFlicSprite() { delete []FlicFilename; FlicFilename=0; }
  void Discard() { delete Flic; Flic=0; LoopsLeft=NumLoops; }
  void Advance (DWORD deltaTime=0);
  TDib *LockDib (TRenderBuffer &renderBuffer)
    { return LoadFlic()?Flic->GetFrameDib():0; }
  void Render (TRenderBuffer &renderBuffer);
  virtual void Reset() { TSprite::Reset(); LoopsLeft=NumLoops; } 
};


// ====================================
//  TTextSprite CLASS DEFINITION
// ====================================

class TTextSprite : public TSprite
{
protected:
  TDib *Dib;			// rendered text
  char far *Text;		// text to be displayed
  uint16 DrawTextFlags;		// DrawText flags
  TColor TextColor;		// foreground text color
  TFont *TextFont;		// font used to render text (optional)
  bool AutoDeleteFont;		// delete font automatically?
  bool AutoStop;		// stop automatically when discarded
  void CalcSpriteSize();
  void DeleteFont();
public:
  TTextSprite (const char far *text, int width, TColor textColor=TColor::White,
    uint16 dtFlags=DT_WORDBREAK|DT_CENTER|DT_NOPREFIX);
  ~TTextSprite() { delete []Text; Text=0; DeleteFont(); }
  virtual void Discard();
  inline bool GetAutoStop() { return AutoStop; }
  inline int GetDrawTextFlags() { return DrawTextFlags; }
  inline TFont *GetFont() { return TextFont; }
  inline TColor GetTextColor() { return TextColor; }
  TDib *LockDib (TRenderBuffer &renderBuffer);
  inline void SetAutoStop (bool autoStop) { AutoStop=autoStop; }
  inline void SetDrawTextFlags (uint16 dtFlags) { DrawTextFlags=dtFlags; }
  inline void SetFont (TFont *font, bool autoDeleteFont=false);
  void SetText (const char far *text, int width);
  inline void SetTextColor (TColor textColor) { TextColor=textColor; }
};


// ====================================
//  TMovieSprite CLASS DEFINITION
// ====================================

//
// Credit flags
//

#define MSCF_HEADING		0x0001	// use heading font and color?
#define MSCF_ENDING		0x0002	// stop rendering
#define MSCF_WHITESPACE		0x0004
#define MSCF_JUSTIFY_LEFT	0x0010	// left justify
#define MSCF_JUSTIFY_CENTER	0x0000	// center text (default)
#define MSCF_JUSTIFY_RIGHT	0x0020	// right justify
#define MSCF_JUSTIFY_ANY	0x00F0	// justify bitmask

class TMovieSprite : public TSpriteGroup
{
protected:
  TFont *HeadingFont;		// heading font
  TFont *NormalFont;		// normal font
  TColor HeadingColor;		// heading color
  TColor NormalColor;		// normal color
  int RollPosition;		// trailer position
public:
  TMovieSprite (TSize size);
  ~TMovieSprite();
  void Render (TRenderBuffer &renderBuffer);
  inline void ResetRoll() { RollPosition=Size.cy; }
  TSprite *Roll (TSprite *sprite, uint16 creditFlags=0);
  TDibSprite *RollDib (TDib *dib, uint16 creditFlags=0);
  TDibSprite *RollDib (TModule *module, uint dibId, uint16 creditFlags=0)
    { return RollDib(new TDib(*module,dibId),creditFlags); }
  TTextSprite *RollText (const char far *text, uint16 creditFlags=0);
  TTextSprite *RollText (uint textId, uint16 creditFlags=0)
    { return RollText(GetApplicationObject()->LoadString(textId).c_str(),creditFlags); }
  inline void SetHeadingColor (TColor headingColor) { HeadingColor=headingColor; }
  inline void SetNormalColor (TColor normalColor) { NormalColor=normalColor; }
};


// ====================================
//  TMovie CLASS DEFINITION
// ====================================

class TMovie : public TWindow
{
protected:
  int TimerId;			// timer used to update sprites
  TSprite *Sprite;		// movie sprite (usually TSpriteGroup)
  TRenderBuffer *RenderBuffer;	// rendering buffer
  bool ClearBackground;		// clear background before renders?
  void CleanupWindow();
  void DrawFrame (TDC &dc);
  void EvTimer (uint timerId);
  char *GetClassName() { return "TMovie"; }
  void Paint (TDC &dc, bool erase, TRect &rect);
  virtual void Render();
  void SetupWindow();
public:
  TMovie (TWindow *parent, int id, int x, int y, TSize frameSize, TSprite *sprite, TModule *module=0);
  TMovie (TWindow *parent, int resourceId, TSize frameSize, TSprite *sprite, TModule *module=0);
  virtual ~TMovie();
  void SetClearBackground (bool clearBackground) { ClearBackground=clearBackground; }
  DECLARE_RESPONSE_TABLE(TMovie);
};


#endif

